@@ -6,12 +6,13 @@ from sqlalchemy import Table, Column, Integer, String, MetaData, Identity, Float
 from sqlalchemy.orm import Session
 import datetime
 from copy import copy
+from pyg_base import logger
 
 _id = '_id'
 _doc = 'doc'
 _root = 'root'
 _deleted = 'deleted'
-
+_archived = 'archived_'
 
 @cache
 def _servers():
@@ -24,7 +25,7 @@ def _servers():
 
 @cache
 def _schema():
-    return cfg_read().get('sql_schema', None)
+    return cfg_read().get('sql_schema', 'dbo')
 
 @cache
 def _database():
@@ -75,10 +76,11 @@ def _pairs2connection(*pairs, **connection):
     for pair in pairs:
         ps = pair.split(';')
         for p in ps:
-            k, v = p.split('=')
-            k = k.strip()
-            v = v.strip().replace("'","")
-            connection[k] = v
+            if p:
+                k, v = p.split('=')
+                k = k.strip()
+                v = v.strip().replace("'","")
+                connection[k] = v
     connection = {k.lower() : v.replace(' ','+').replace('{','').replace('}','') for k, v in connection.items() if v is not None}
     return connection
 
@@ -102,6 +104,19 @@ def get_cstr(*pairs, **connection):
         params = '&'.join('%s=%s'%(k,v) for k,v in connection.items())
         return 'mssql+pyodbc://%(server)s/%(db)s%(params)s'%dict(server=server, db = db, params = '?' +params if params else '')
 
+def create_schema(engine, schema):
+    if schema is None:
+        return
+    try:
+        if schema not in engine.dialect.get_schema_names(engine):
+            engine.execute(sa.schema.CreateSchema(schema))
+            logger.info('creating schema: %s'%schema)
+    except AttributeError: #MS SQL vs POSTGRES
+        if not engine.dialect.has_schema(engine, schema):
+            engine.execute(sa.schema.CreateSchema(schema))
+            logger.info('creating schema: %s'%schema)
+    return schema
+    
 def get_engine(*pairs, **connection):    
     """
     returns a sqlalchemy engine object
@@ -119,7 +134,7 @@ def get_engine(*pairs, **connection):
     try:
         sa.inspect(e)
     except Exception:
-        print('creating db... ', db)
+        logger.info('creating database: %s'%db)
         create_database(cstr)
         e = sa.create_engine(cstr)       
     return e
@@ -135,7 +150,6 @@ _types = {str: String, 'str' : String,
 
 _orders = {1 : asc, True: asc, 'asc': asc, asc : asc, -1: desc, False: desc, 'desc': desc, desc: desc}
 
-
 def sql_table(table, db = None, non_null = None, nullable = None, _id = None, schema = None, server = None, reader = None, writer = None, pk = None, doc = None, mode = None):
     """
     Creates a sql table. Can also be used to simply read table from the db
@@ -181,13 +195,13 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
     if isinstance(table, str):
         values = table.split('.')
         if len(values) == 2:
-            db = db or values[0]
-            if db != values[0]:
-                raise ValueError('db cannot be both %s and %s'%(values[0], db))
+            schema = schema or values[0]
+            if schema != values[0]:
+                raise ValueError('schema cannot be both %s and %s'%(values[0], db))
             table = values[1]
         elif len(values)>2:
             raise ValueError('not sure how to translate this %s into a db.table format'%table)
-    e = get_engine(server = server, db = db)
+    e = get_engine(server = server, db = db, schema = schema)
     
     non_null = non_null or {}
     nullable = nullable or {}
@@ -209,8 +223,9 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
         table_name = table 
     else:
         table_name = table.name
-        schema = table.schema
-    schema = schema or _schema()
+        if schema is None:
+            schema = table.schema
+    schema = create_schema(e, schema or _schema())
     if doc is True:
         doc = _doc
     meta = MetaData()
@@ -239,13 +254,15 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
         col_names = [col.name for col in cols]
         non_nulls = [Column(k, _types.get(t, t), nullable = False) for k, t in pks.items() if k not in col_names]
         nullables = [Column(k.lower(), _types.get(t, t)) for k, t in nullable.items() if k not in col_names] 
-        docs = [Column(doc, String, nullable = True)] if doc is not None else []
+        docs = [Column(doc, String, nullable = True)] if doc else []
         cols = cols + non_nulls + nullables + docs
         if len(cols) == 0:
             raise ValueError('You seem to be trying to create a table with no columns? Perhaps you are trying to point to an existing table and getting its name wrong?')
-        tbl = Table(table_name, meta, *cols)
+        logger.info('creating table: %s.%s.%s%s'%(db, schema, table_name, [col.name for col in cols]))
+        tbl = Table(table_name, meta, *cols, schema = schema)
         meta.create_all(e)
     else:
+        # logger.info('schema :%s'%schema)
         tbl = Table(table_name, meta, autoload_with = e, schema = schema)
         cols = tbl.columns
         non_nulls = [Column(k, _types.get(t, t), nullable = False) for k, t in pks.items()]
@@ -848,7 +865,7 @@ class sql_cursor(object):
         
         columns = columns or self.columns
         edoc = self._enrich(doc, columns)
-        if self.doc is None or isinstance(edoc.get(self.doc), dict):
+        if self.doc is None or self.doc is False or isinstance(edoc.get(self.doc), dict):
             return edoc
         else:
             edoc = {key: value for key, value in edoc.items() if key in columns}
@@ -888,7 +905,7 @@ class sql_cursor(object):
         """
         converts a document which is of the format {self.doc : doc} into a regular looking document
         """
-        if self.doc is None or not isinstance(doc.get(self.doc), dict):
+        if self.doc is None or self.doc is False or not isinstance(doc.get(self.doc), dict):
             return Dict(doc) if type(doc) == dict else doc
         res = doc[self.doc]
         columns = columns or self.columns
@@ -1238,21 +1255,22 @@ class sql_cursor(object):
         return res
 
     def _is_deleted(self):
-        return self.db.startswith('deleted_')
+        return is_str(self.schema) and self.schema.startswith(_archived)
 
     @property
     def deleted(self):
         if self._is_deleted():
             return self
         else:        
-            db_name = 'deleted_' + self.db
-            res = sql_table(table = self.table, db = db_name, non_null = dict(deleted = datetime.datetime), 
+            schema = _archived + (self.schema or '')
+            # logger.info('archived schema: %s'%schema)
+            res = sql_table(table = self.table, db = self.db, non_null = dict(deleted = datetime.datetime), 
                             server = self.server, 
                             pk = self.pk, 
                             doc = self.doc, 
                             writer = self.writer, 
                             reader = self.reader, 
-                            schema = self.schema)
+                            schema = schema)
             res.spec = self.spec
             res.order = self.order
             res.selection = self.selection
