@@ -8,15 +8,34 @@ sql_table_ = cache(sql_table)
 _sql = '.sql'
 _dictable = '.dictable'
 _dictable_decoded = encode(dictable_decoded)
+_key = 'key'
+_data = 'data'
 
-def path_to_connection(path):
+def sql_binary_store(path):
+    """
+    splits a path which resembles a sql-alchemy connection string, to its bits
+
+    Parameters
+    ----------
+    path : str
+        extended connection string.
+
+    Returns
+    -------
+    dict
+        various connection parameters. specifically, the cursor parameter actually generates the table
+    """
     u, p = path.split('//')
     ps = p.split('/')
     server = ps[0]
     params = ps[1]
-    db, prm = params.split('?')
-    connections = _pairs2connection(*prm.split('&'))
-    doc = connections.get('doc', 'true')
+    if '?' in params:
+        db, prm = params.split('?')
+        connections = _pairs2connection(*prm.split('&'))
+    else:
+        db = params
+        connections = {}
+    doc = connections.pop('doc', 'true')
     doc = dict(true = True, false = False).get(doc.lower(), doc)        
     url = '%s//%s/%s'%(u, server, params)
     table = ps[2].split('/')[0]
@@ -26,56 +45,93 @@ def path_to_connection(path):
     else:
         schema = None
     schema = connections.pop('schema', schema)
-    cursor = sql_table_(table = table, db = db, schema = schema, pk = 'key', 
-                        non_null = dict(value = bin), doc = doc)
+    cursor = sql_table(table = table, db = db, schema = schema, pk = _key, 
+                        non_null = {_data : bin}, doc = doc)
     connections.update(dict(url = url, server = server, schema  = schema, cursor = cursor, root = root, table = table))
     return Dict(connections)
 
 
-def sql_dumps(df, path):
+def sql_dumps(obj, path):
     """
-    path = 'mssql+pyodbc://localhost/test_db?driver=ODBC+Driver+17+for+SQL+Server&schema=xyz'
-
+    >>> from pyg import *
+    >>> path = 'mssql+pyodbc://localhost/test_db/xxx.test_table/key'
+    >>> self = sql_binary_store(path).cursor
+    >>> self.deleted
+    >>> obj = pd.Series([1,2,3])
+    >>> sql_dumps(obj, path)
+    >>> sql_loads(path)
     Parameters
     ----------
-    df : TYPE
-        DESCRIPTION.
-    path : TYPE
-        DESCRIPTION.
+    obj : object
+        item to be pickled into binary.
+    path : str
+        path sqlalchemy-like to save the pickled binary in.
 
     Returns
     -------
-    None.
+    string 
+        path
 
     """
-    res = path_to_connection(path)
-    value = pickle.dumps(df)
+    res = sql_binary_store(path)
+    data = pickle.dumps(obj)
     cursor = res.cursor
     root = res.root
     # print('dumping into...\n', cursor)
-    cursor.update_one(dict(key = root, value = value))
+    cursor.update_one({_key : root, _data : data})
     # print(cursor)
     return path
 
 def sql_loads(path):
-    res = path_to_connection(path)
+    res = sql_binary_store(path)
     cursor = res.cursor
     root = res.root
-    row = cursor.inc(**dict(key = root))
+    row = cursor.inc(**{_key :root})
     if len(row) == 0:
         # print('no documents found in...\n', row)
         raise ValueError('no document found in %s' %(res-'cursor'))
+    elif len(row) > 1:
+        raise ValueError('multiple documents found \n%s'%row)
     else:
         # print('loading from...\n', row)
-        value = row[0]['value']
-    df = pickle.loads(value)
-    return df
+        data = row[0][_data]
+        obj = pickle.loads(data)
+        return obj
     
 _sql_loads = encode(sql_loads)
 
 def sql_encode(value, path):
     """
     encodes a single DataFrame or a document containing dataframes into a an abject of multiple pickled files that can be decoded
+
+    Parameters:
+    ----------
+    value : document or dataframe
+        value to be encoded inside a sql database
+        
+    path : str
+        a sqlalchemy-like string
+        
+    Example: writing a single dataframe
+    --------
+    >>> from pyg import * 
+    >>> value = pd.Series([1,2])
+    >>> path = 'mssql+pyodbc://localhost/database_here?doc=false/xyz.table_name/root_of_doc'
+    >>> res = sql_encode(value, path)
+    >>> table = sql_table(db = 'database_here', schema = 'xyz', table = 'table_name')
+    >>> assert len(table.inc(key = 'root_of_doc'))>0
+    >>> sql_loads(path)
+    
+    Example: writing a document
+    ---------------------------
+    >>> from pyg import * 
+    >>> value = dict(a = pd.Series([1,2]), b = pd.Series([3,4]))
+    >>> path = 'mssql+pyodbc://localhost/database_here?doc=false/xyz.table_name/root_of_doc'
+    >>> res = sql_encode(value, path)
+    >>> table = sql_table(db = 'database_here', schema = 'xyz', table = 'table_name')
+    >>> keys = table.distinct('key')
+    >>> assert 'root_of_doc/a' in keys and 'root_of_doc/b' in keys    
+    >>> assert eq(value['a'],sql_loads(path+'/a'))
     """
     if path.endswith(_sql):
         path = path[:-len(_sql)]
@@ -83,7 +139,7 @@ def sql_encode(value, path):
         path = path[:-1]
     if is_pd(value) or is_arr(value):
         path = root_path_check(path)
-        return dict(_obj = _sql_loads, path = sql_dumps(value, path + _sql))       
+        return dict(_obj = _sql_loads, path = sql_dumps(value, path))       
     elif is_dict(value):
         res = type(value)(**{k : sql_encode(v, '%s/%s'%(path,k)) for k, v in value.items()})
         if isinstance(value, dictable):
@@ -109,7 +165,7 @@ def sql_write(doc, root = None):
                      db = 'bbgs', 
                      pk = ['ticker', 'item'], 
                      server = 'localhost', 
-                     writer = 'mssql+pyodbc://localhost/bbgs?driver=ODBC+Driver+17+for+SQL+Server/dbo.bbg_data/%ticker/%item.sql', 
+                     writer = 'mssql+pyodbc://localhost/bbgs?driver=ODBC+Driver+17+for+SQL+Server&doc=false/bbg_data/%ticker/%item.sql', 
                      doc = True)
     >>> path = db().writer
     >>> res = path_to_connection(path)
